package answer

import (
	"answer/pkg/cache"
	"answer/pkg/search"
	"context"
	"errors"
	"fmt"
	go_llm "gitea.stevedudenhoeffer.com/steve/go-llm"
	"io"
	"log/slog"
	"net/http"
	"net/url"
	"strings"
)

var ErrMaxTries = errors.New("maximum number of pages tried reached")
var ErrMaxAnswers = errors.New("maximum number of answers parsed reached")
var ErrTooManyArguments = errors.New("too many arguments")

type Question struct {
	// Question is the question to answer
	Question string

	Model go_llm.ChatCompletion

	Search search.Search

	Cache cache.Cache
}

// Answers is a list of answers to a question
type Answers []string

type Options struct {
	// MaxSearches is the maximum possible number of searches to execute for this question.  If this is set to 5, the function could
	// search up to 5 possible times to find an answer.
	MaxSearches int

	// MaxThinks is the maximum number of times to think about a question.  A "Think" is different than a search in that
	// the LLM just breaks the question down into smaller parts and tries to answer them.  This is useful for complex
	// questions that are hard to answer since LLMs are better at answering smaller questions.
	MaxThinks int

	// MaxTries is the absolute maximum number of pages to try to get an answer from.  For instance, if MaxSearches is 5 and
	// 5 pages are tried and no answers are found, the function will return ErrMaxTries.
	MaxTries int
}

var DefaultOptions = Options{
	MaxSearches: 5,
	MaxThinks:   10,
	MaxTries:    5,
}

type Result struct {
	Result string
	Error  error
}

func fanExecuteToolCalls(ctx context.Context, toolBox *go_llm.ToolBox, calls []go_llm.ToolCall) []Result {
	var results []Result
	var resultsOutput = make(chan Result, len(calls))

	fnCall := func(call go_llm.ToolCall) Result {
		str, err := toolBox.Execute(ctx, call)
		if err != nil {
			return Result{
				Error: err,
			}
		}

		return Result{
			Result: str,
		}
	}

	for _, call := range calls {
		go func(call go_llm.ToolCall) {
			resultsOutput <- fnCall(call)
		}(call)
	}

	for i := 0; i < len(calls); i++ {
		result := <-resultsOutput
		results = append(results, result)
	}

	close(resultsOutput)

	return results
}

type article struct {
	URL   string
	Title string
	Body  string
}

func extractArticle(ctx context.Context, u *url.URL) (res article, err error) {
	defer func() {
		e := recover()

		if e != nil {
			if e, ok := e.(error); ok {
				err = fmt.Errorf("panic: %w", e)
			} else {
				err = fmt.Errorf("panic: %v", e)
			}
		}
	}()

	req, err := http.NewRequestWithContext(ctx, "GET", u.String(), nil)

	if err != nil {
		return res, fmt.Errorf("error creating request: %w", err)
	}

	resp, err := c.cl.Do(req)

	if err != nil {
		return res, fmt.Errorf("error getting response: %w", err)
	}

	defer func(Body io.ReadCloser) {
		err := Body.Close()
		if err != nil {
			slog.Error("error closing body", "error", err)
		}
	}(resp.Body)

	if resp.StatusCode < 200 || resp.StatusCode >= 300 {
		return "", fmt.Errorf("bad response: %d: %s", resp.StatusCode, resp.Status)
	}

	b, err := io.ReadAll(resp.Body)

	if err != nil {
		return "", fmt.Errorf("error reading body: %w", err)
	}

	g := goose.New()
	article, err := g.ExtractFromRawHTML(string(b), target)

	if err != nil {
		return "", fmt.Errorf("error extracting article: %w", err)
	}

	return article.CleanedText, nil
	panic("not implemented")
	return article{}, nil
}

func doesTextAnswerQuestion(ctx context.Context, q Question, text string) (string, error) {
	fnAnswer := go_llm.NewFunction(
		"answer",
		"The answer from the given text that answers the question.",
		func(ctx context.Context, args struct {
			Answer string `description:"the answer to the question, the answer should come from the text"`
		}) (string, error) {
			return args.Answer, nil
		})

	fnNoAnswer := go_llm.NewFunction(
		"no_answer",
		"Indicate that the text does not answer the question.",
		func(ctx context.Context, args struct{}) (string, error) {
			return "", nil
		})

	req := go_llm.Request{
		Messages: []go_llm.Message{
			{
				Role: go_llm.RoleSystem,
				Text: "Evaluate the given text to see if it answers the question from the user.  The text is as follows:",
			},
			{
				Role: go_llm.RoleSystem,
				Text: text,
			},
			{
				Role: go_llm.RoleUser,
				Text: q.Question,
			},
		},
		Toolbox: go_llm.NewToolBox(fnAnswer, fnNoAnswer),
	}

	res, err := q.Model.ChatComplete(ctx, req)

	if err != nil {
		return "", err
	}

	if len(res.Choices) == 0 {
		return "", nil
	}

	if len(res.Choices[0].Calls) == 0 {
		return "", nil
	}

	return req.Toolbox.Execute(ctx, res.Choices[0].Calls[0])
}

func functionSearch(ctx context.Context, q Question, searchTerm string) (string, error) {
	res, err := q.Search.Search(ctx, searchTerm)
	if err != nil {
		return "", err
	}

	if len(res) == 0 {
		return "", nil
	}

	// first pass try to see if any provide the result without needing archive
	for _, r := range res {
		trimmed := strings.TrimSpace(r.URL)
		if trimmed == "" {

		}
		u, err := url.Parse(trimmed)
		if err != nil {
			continue
		}

		a, err := extractArticle(ctx, u)

		if err != nil {
			continue
		}

		if a.Title != "" && a.Body != "" {
			answer, err := doesTextAnswerQuestion(ctx, q, a.Body)

			if err != nil {
				slog.Error("error checking if text answers question", "question", q.Question, "error", err)
				continue
			}

			if answer != "" {
				return answer, nil
			}
		}
	}

	return "", nil
}

func functionThink(ctx context.Context, q Question) (string, error) {
	fnAnswer := go_llm.NewFunction(
		"answer",
		"Answer the question.",
		func(ctx context.Context, args struct {
			Answer string `description:"the answer to the question"`
		}) (string, error) {
			return args.Answer, nil
		})

	var temp float32 = 0.8
	req := go_llm.Request{
		Messages: []go_llm.Message{
			{
				Role: go_llm.RoleSystem,
				Text: "Answer the given question as accurately and concisely as possible using the answer function.",
			},
			{
				Role: go_llm.RoleUser,
				Text: q.Question,
			},
		},
		Toolbox:     go_llm.NewToolBox(fnAnswer),
		Temperature: &temp,
	}

	res, err := q.Model.ChatComplete(ctx, req)

	if err != nil {
		return "", err
	}

	if len(res.Choices) == 0 {
		return "", nil
	}

	if len(res.Choices[0].Calls) == 0 {
		return "", nil
	}

	return req.Toolbox.Execute(ctx, res.Choices[0].Calls[0])
}

func (o Options) Answer(ctx context.Context, q Question) (Answers, error) {
	fnSearch := go_llm.NewFunction(
		"search",
		"Search the web for an answer to a question.  You can call this function up to "+fmt.Sprint(o.MaxSearches)+" times.",
		func(ctx context.Context, args struct {
			SearchQuery string `description:"what to search the web for for this question"`
			Question    string `description:"what question(s) you are trying to answer with this search"`
		}) (string, error) {
			q2 := q
			q2.Question = args.Question

			return functionSearch(ctx, q2, args.SearchQuery)
		})

	fnThink := go_llm.NewFunction(
		"think",
		"Think about a question.  This is useful for breaking down complex questions into smaller parts that are easier to answer.",
		func(ctx context.Context, args struct {
			Question string `description:"the question to think about"`
		}) (string, error) {
			q2 := q
			q2.Question = args.Question

			return functionThink(ctx, q2)
		})

	fnAnswer := go_llm.NewFunction(
		"answer",
		"You definitively answer a question, if you call this it means you know the answer and do not need to search for it or use any other function to find it",
		func(ctx context.Context, args struct {
			Answer string `description:"the answer to the question"`
		}) (string, error) {
			return args.Answer, nil
		})

	var funcs = []*go_llm.Function{fnAnswer}

	if o.MaxSearches > 0 {
		funcs = append(funcs, fnSearch)
	}

	if o.MaxThinks > 0 {
		funcs = append(funcs, fnThink)
	}

	var temp float32 = 0.8

	req := go_llm.Request{
		Messages: []go_llm.Message{
			{
				Role: go_llm.RoleSystem,
				Text: "You are being asked to answer a question.  You must respond with a function.  You can answer it if you know the answer, or if some functions exist you can use those to help you find the answer.",
			},
			{
				Role: go_llm.RoleUser,
				Text: q.Question,
			},
		},
		Toolbox:     go_llm.NewToolBox(funcs...),
		Temperature: &temp,
	}

	res, err := q.Model.ChatComplete(ctx, req)

	if err != nil {
		return nil, err
	}

	if len(res.Choices) == 0 {
		return nil, nil
	}

	if len(res.Choices) > o.MaxSearches {
		res.Choices = res.Choices[:o.MaxSearches]
	}

	var answers []QuestionAnswer
	choicesOutput := make(chan QuestionAnswer, len(res.Choices))

	for _, choice := range res.Choices {
		fnChoice := func(choice go_llm.ResponseChoice) QuestionAnswer {
			var calls []CallResult
			var callsOutput = make(chan CallResult, len(choice.Calls))
			fnCall := func(call go_llm.ToolCall) CallResult {
				str, err := req.Toolbox.Execute(ctx, call)

				if err != nil {
					return CallResult{
						Error: err,
					}
				}

				return CallResult{
					Result: str,
				}
			}

			for _, call := range choice.Calls {
				go func(call go_llm.ToolCall) {
					callsOutput <- fnCall(call)
				}(call)
			}

			for i := 0; i < len(choice.Calls); i++ {
				result := <-callsOutput
				calls = append(calls, result)
			}

			close(callsOutput)

		}

	}

}

func Answer(ctx context.Context, q Question) (Answers, error) {
	return DefaultOptions.Answer(ctx, q)
}
